#include <config.h>

#include <cmath>
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>
#include <dune/common/shared_ptr.hh>
#include <dune/common/parametertree.hh>
#include <dune/common/parametertreeparser.hh>
#include <dune/istl/matrix.hh>
#include <dune/common/timer.hh>

// dune-istl includes
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/preconditioners.hh>

// dune-fufem includes
#include <dune/fufem/functionspacebases/p1nodalbasis.hh>
#include <dune/fufem/assemblers/assembler.hh>
#include <dune/fufem/assemblers/localassemblers/laplaceassembler.hh>
#include <dune/fufem/assemblers/localassemblers/l2functionalassembler.hh>
#include <dune/fufem/boundarypatch.hh>
#include <dune/fufem/functiontools/boundarydofs.hh>

// dune-grid includes
#include <dune/grid/uggrid.hh>
#include <dune/grid/io/file/vtk/vtkwriter.hh>
#include <dune/grid/utility/structuredgridfactory.hh>
#include <dune/grid/common/gridfactory.hh>
#include <dune/grid/common/mcmgmapper.hh>

// dune-solver includes
#include <dune/solvers/norms/energynorm.hh>
#include <dune/solvers/norms/twonorm.hh>
//#include <dune/solvers/norms/onenorm.hh>
#include <dune/faultnetworks/compressedmultigridtransfer.hh>

#include <dune/faultnetworks/utils/matrixreader.hh>
#include <dune/faultnetworks/utils/vectorreader.hh>
#include <dune/faultnetworks/oscrhs.hh>

#include <dune/faultnetworks/preconditioners/faultpreconditioner.hh>



#include <dune/faultnetworks/solvers/osccgsolver.hh>

#include <dune/faultnetworks/faultp1nodalbasis.hh>
#include <dune/faultnetworks/levelinterfacenetwork.hh>
#include <dune/faultnetworks/faultinterface.hh>
#include <dune/faultnetworks/assemblers/interfaceoperatorassembler.hh>
#include <dune/faultnetworks/assemblers/intersectionjumpmassassembler.hh>
#include <dune/faultnetworks/assemblers/osclocalassembler.hh>
#include <dune/faultnetworks/faultfactories/oscunitcube.hh>
#include <dune/faultnetworks/faultfactories/twofaultfactory.hh>
#include <dune/faultnetworks/dgmgtransfer.hh>
#include <dune/faultnetworks/utils/debugutils.hh>

#include <dune/istl/superlu.hh>
#include <dune/istl/umfpack.hh>

const int dim = 2;
typedef double ctype;


typedef typename Dune::FieldVector<ctype, dim> VectorType;
typedef typename Dune::BitSetVector<1> BitVector;
typedef typename Dune::BCRSMatrix<Dune::FieldMatrix<ctype, 1, 1>> MatrixType;


#if HAVE_UG
typedef typename Dune::UGGrid<dim> GridType;
typedef OscUnitCube<GridType, 2> GridOb;
#else 
#error No UG!
#endif


typedef typename GridType::LevelGridView GV;

typedef P1NodalBasis<GV, ctype> Basis;
typedef typename Basis::LocalFiniteElement FE;

typedef FaultP1NodalBasis<GV, ctype> DGBasis;
typedef typename DGBasis::LocalFiniteElement DGFE;


/*
 void print(const Dune::BlockVector<Dune::FieldVector<ctype, 1>>& x, std::string message){
        std::cout << message << std::endl;
        for (size_t i=0; i<x.size(); i++) {
            std::cout << x[i] << " ";
        }
        std::cout << std::endl << std::endl;
    }

    void print(const MatrixType& mat, std::string message){
        std::cout << message << std::endl;
        for (size_t i=0; i<mat.N(); i++) {
            for (size_t j=0; j<mat.M(); j++) {
                if (mat.exists(i,j))
                    std::cout << mat[i][j] << " ";
                else 
                    std::cout << "0 ";
            }
            std::cout << std::endl;
        }
        std::cout << std::endl;
    }*/

    // res = A + B
    void add(MatrixType& A, MatrixType& B, MatrixType& res) {
          if (A.N()!=B.N() || A.M()!=B.M())
              DUNE_THROW(RangeError, "faultnetworks::add() Matrices must have same dimensions!");

          typedef typename std::map<std::pair<size_t, size_t>, ctype> MatEntriesType;
          typedef typename MatEntriesType::iterator MatEntriesIterator;
          typedef typename MatrixType::row_type RowType;
          typedef typename RowType::ConstIterator ColumnIterator;

          MatEntriesType matEntries;
          MatEntriesIterator entryIt;

          MatrixIndexSet joinedIdxSet(A.N(), A.M());
          joinedIdxSet.import(A);

          for(size_t rowIdx = 0; rowIdx<B.N(); rowIdx++) {
              const RowType& BRow = B[rowIdx];
              const RowType& ARow = A[rowIdx];

              ColumnIterator cIt    = BRow.begin();
              ColumnIterator cEndIt = BRow.end();

              for(; cIt!=cEndIt; ++cIt) {
                  const size_t colIdx = cIt.index();

                  if (A.exists(rowIdx, colIdx)) {
                      matEntries[std::make_pair(rowIdx, colIdx)] = ARow[colIdx] + BRow[colIdx];
                  } else {
                      matEntries[std::make_pair(rowIdx, colIdx)] = BRow[colIdx];
                      joinedIdxSet.add(rowIdx, colIdx);
                  }
              }
          }

          joinedIdxSet.exportIdx(res);

          for(size_t rowIdx = 0; rowIdx<res.N(); rowIdx++) {
              RowType& row = res[rowIdx];
              const RowType& ARow = A[rowIdx];

              ColumnIterator cIt    = row.begin();
              ColumnIterator cEndIt = row.end();

              for(; cIt!=cEndIt; ++cIt) {
                  const size_t colIdx = cIt.index();
                  entryIt = matEntries.find(std::make_pair(rowIdx, colIdx));

                  if (entryIt==matEntries.end()) {
                      row[colIdx] = ARow[colIdx];
                  } else {
                      row[colIdx] = entryIt->second;
                  }
              }
          }
      }


int main(int argc, char** argv) try
{
 
  // parse parameter file
    ParameterTree parameterSet;
    if (argc==2)
        ParameterTreeParser::readINITree(argv[1], parameterSet);
    else
        ParameterTreeParser::readINITree("/home/mi/podlesny/software/dune/faultnetworks/src/uniformfaultnetworks/2faultscontcoarse2level.parset", parameterSet);

    const std::string path = parameterSet.get<std::string>("path");
    const std::string resultPath = parameterSet.get<std::string>("resultPath");

    int problemCount = 0;
    while (parameterSet.hasSub("problem" + std::to_string(problemCount))) {
	
    ParameterTree& problemParameters = parameterSet.sub("problem" + std::to_string(problemCount));

    const std::string oscDataFile = problemParameters.get<std::string>("oscDataFile");

    const int coarseResolution = problemParameters.get<int>("coarseResolution");
    const int fineResolution = problemParameters.get<int>("fineResolution");
    const int exactResolution = problemParameters.get<int>("exactResolution");

    const int faultDistance = problemParameters.get<int>("faultDistance");

    const int maxIterations = problemParameters.get<int>("maxIterations");
    const double solverTolerance = problemParameters.get<double>("tol");

    //const int coarseGridN = std::pow(2, coarseResolution);
    const int fineGridN = std::pow(2, fineResolution);

   // std::ofstream out(resultPath + oscDataFile + "_amp" + std::to_string((int) amplitudeExp) + "_" + std::to_string(coarseResolution) + "_" + std::to_string(fineResolution) + "_" + std::to_string(exactResolution) + ".log");

   // TODO: after debugging: use log to file again

    std::ofstream out(resultPath + oscDataFile + "_" + std::to_string(coarseResolution) + "_" + std::to_string(fineResolution) + "_" + std::to_string(exactResolution) + "_" + std::to_string(faultDistance) + ".log");
    std::streambuf *coutbuf = std::cout.rdbuf(); //save old buf
    std::cout.rdbuf(out.rdbuf()); //redirect std::cout to out.txt!


    std::cout << std::endl;
    std::cout << "Parameter file read successfully!" << std::endl;

    typedef Dune::Matrix<Dune::FieldMatrix<ctype,1,1>, std::allocator<Dune::FieldMatrix<ctype,1,1>>> OscMatrixType;
    OscMatrixType matrix;

    // interface integral jump penalty, B:
    OscMatrixType B(2,2);
    B[0][0] = 1;
    B[1][1] = 1;
    B[0][1] = -1;
    B[1][0] = -1;

    MatrixReader<OscMatrixType> matrixReader(path + oscDataFile);
    matrixReader.read(matrix);
	
    /*
    for (size_t i=0; i<matrix.N(); i++) {
	for (size_t j=0; j<matrix.M(); j++) {
		matrix[i][j] = 1 + matrix[i][j]*(amplitude - 1);
	}
    }*/

    const int oscGridN = matrix.N();

    if (oscGridN>fineGridN)
	DUNE_THROW(Dune::Exception, "Provided oscData too large!");

    std::cout << "OscData file read successfully!" << std::endl << std::endl;

    TwoFaultFactory<GridType> faultFactory(faultDistance, coarseResolution, fineResolution, exactResolution);
    const GridType& grid = faultFactory.grid();
    const LevelInterfaceNetwork<GV>& coarseFaultNetwork = faultFactory.coarseInterfaceNetwork();
    const LevelInterfaceNetwork<GV>& fineFaultNetwork = faultFactory.fineInterfaceNetwork();
    const LevelInterfaceNetwork<GV>& exactFaultNetwork = faultFactory.exactInterfaceNetwork();

    const int fineLevelIdx = fineResolution - coarseResolution;
    const int exactLevelIdx = exactResolution - coarseResolution;

/*
    GV gridView(grid.levelGridView(0));


    VTKWriter<GV> vtkwriter(gridView);
    vtkwriter.pwrite("coarseGrid", "/home/mi/podlesjo/data/", "/home/mi/podlesjo/data/");//("~/home/mi/podlesjo/data/coarseGrid");

    GV gridViewFine(grid.levelGridView(fineLevelIdx));


    VTKWriter<GV> vtkwriterFine(gridViewFine);
    vtkwriterFine.pwrite("fineGrid", "/home/mi/podlesjo/data/", "/home/mi/podlesjo/data/");//("~/home/mi/podlesjo/data/coarseGrid");
*/
    typedef  typename GV::Codim <0>::Iterator  ElementLevelIterator;
    typedef  typename ElementLevelIterator::Entity::Geometry  LevelGeometry;
    int i,j;
    
    Timer totalTimer;
    totalTimer.reset();
    totalTimer.start();
 // ----------------------------------------------------------------
    // ---              compute initial iterate                     ---
    // ----------------------------------------------------------------
    std::cout << std::endl;
    std::cout << "Computing initial iterate!" << std::endl;
 
    GV coarseGridView(grid.levelGridView(0));
    GV fineGridView(grid.levelGridView(fineLevelIdx));
    DGBasis fineBasis(fineGridView, fineFaultNetwork);

    std::cout << "Setting coarse grid oscData:" << std::endl;

    Dune::LevelMultipleCodimMultipleGeomTypeMapper <GridType,  Dune::MCMGElementLayout > fineMapper(grid, fineLevelIdx);

    // allocate  a  vector  for  the  data
    std::vector<ctype> fineOscData(fineMapper.size());

    ElementLevelIterator endIt = fineGridView.end <0>();
    for (ElementLevelIterator  it = fineGridView.begin <0>(); it!=endIt; ++it) {
        const LevelGeometry  geometry = it->geometry();

        Dune::FieldVector <ctype, dim > global = geometry.center();
        i = oscGridN - std::ceil(global[0]*oscGridN);
        j = std::floor(global[1]*oscGridN);
        fineOscData[fineMapper.index(*it)] = matrix[i][j];

        //std::cout << "elem: " << (int) fineMapper.map(*it) << "| (" << global[0] << ", " << global[1] << "), " << "(" << i+1 << ", " << j+1 << "), " << fineOscData[fineMapper.map(*it)] << std::endl;
    }


    Dune::LevelMultipleCodimMultipleGeomTypeMapper <GridType,  Dune::MCMGElementLayout > coarseMapper(grid, 0);

    // allocate  a  vector  for  the  data
    std::vector<ctype> coarseOscData(coarseMapper.size());
 
    typedef HierarchicLevelIterator<GridType> HierarchicLevelIteratorType;

    typedef  typename GV::Codim <0>::Iterator  ElementLevelIterator;
    ElementLevelIterator endElemIt = coarseGridView.end <0>();
    for (ElementLevelIterator  it = coarseGridView.begin <0>(); it!=endElemIt; ++it) {
        int counter = 0;
	ctype sum = 0;
 	  
	HierarchicLevelIteratorType endHierIt(grid, *it, HierarchicLevelIteratorType::PositionFlag::end, fineLevelIdx);
	for (HierarchicLevelIteratorType hierIt(grid, *it, HierarchicLevelIteratorType::PositionFlag::begin, fineLevelIdx); hierIt!=endHierIt; ++hierIt) {
		sum += fineOscData[fineMapper.index(*hierIt)];     
		counter++;
	}
	coarseOscData[coarseMapper.index(*it)] = sum/counter;
	//std::cout << "elem: " << (int) mapper.map(*it) << "| "<< coarseOscData[coarseMapper.map(*it)] << std::endl;
    } 
       
    // assemble stiffness matrix
    Basis coarseBasis(coarseGridView);

    MatrixType coarseStiffMat;
    Assembler<Basis, Basis> coarseAssembler(coarseBasis, coarseBasis);
    OscLocalAssembler<GridType, FE, FE> coarseLocalAssembler(coarseOscData, coarseMapper);
    coarseAssembler.assembleOperator(coarseLocalAssembler, coarseStiffMat);

    // assemble rhs
    Dune::BlockVector<Dune::FieldVector<double,1>> coarseRhs, coarseSol, contCoarseSol, initialX;
    coarseRhs.resize(coarseStiffMat.N());
    coarseSol.resize(coarseStiffMat.M());

    OscRhs<VectorType, Dune::FieldVector<ctype,1>> f;
    L2FunctionalAssembler<GridType, FE> l2FunctionalAssembler(f);
    coarseAssembler.assembleFunctional(l2FunctionalAssembler, coarseRhs);

    typedef typename MatrixType::row_type RowType;
    typedef typename RowType::ConstIterator ColumnIterator;

    // set boundary conditions
    BitVector coarseBoundaryDofs;
    BoundaryPatch<GridType::LevelGridView> coarseBoundaryPatch(coarseGridView, true);
    constructBoundaryDofs(coarseBoundaryPatch, coarseBasis, coarseBoundaryDofs);

    for(size_t i=0; i<coarseBoundaryDofs.size(); i++) {
        if(!coarseBoundaryDofs[i][0])           
		continue;

        const RowType& row = coarseStiffMat[i];

        ColumnIterator cIt    = row.begin();
        ColumnIterator cEndIt = row.end();

        for(; cIt!=cEndIt; ++cIt) {
            coarseStiffMat[i][cIt.index()] = 0;
        }

        coarseStiffMat[i][i]=1;
        coarseRhs[i] = 0;
        coarseSol[i] = 0;
    }


    // compute coarse solution
    Timer timer;
    timer.reset();
    timer.start();
    #if HAVE_SUPERLU
    Dune::InverseOperatorResult res;
    Dune::SuperLU<MatrixType> coarseSolver(coarseStiffMat);
    coarseSolver.apply(coarseSol, coarseRhs, res);
    #else
    #error No SuperLU!
    #endif
    std::cout << "Solving coarse problem: " << timer.elapsed() << " seconds" << std::endl;
    timer.stop();
    timer.reset();	

    CompressedMultigridTransfer<Dune::BlockVector<Dune::FieldVector<ctype,1>>> mgTransfer;
    mgTransfer.setup(coarseGridView, fineGridView);
    mgTransfer.prolong(coarseSol, contCoarseSol);

    //prolong to discontinous fineLevel basis
    fineBasis.prolong(contCoarseSol, initialX);


    // ----------------------------------------------------------------
    // ---              compute exact solution                      ---
    // ----------------------------------------------------------------
    std::cout << std::endl;
    std::cout << "Computing exact solution!" << std::endl;
 
    GV exactGridView(grid.levelGridView(exactLevelIdx));
    DGBasis exactBasis(exactGridView, exactFaultNetwork);


        // attach ocsData to exact grid
        Dune::LevelMultipleCodimMultipleGeomTypeMapper <GridType,  Dune::MCMGElementLayout > exactMapper(grid, exactLevelIdx);

        // allocate  a  vector  for  the  data
        std::vector<ctype> exactOscData(exactMapper.size());

        ElementLevelIterator endIter = exactGridView.end <0>();
        for (ElementLevelIterator  it = exactGridView.begin <0>(); it!=endIter; ++it) {
            const LevelGeometry  geometry = it->geometry();

            Dune::FieldVector <ctype, dim > global = geometry.center();
            i = oscGridN - std::ceil(global[0]*oscGridN);
            j = std::floor(global[1]*oscGridN);
            exactOscData[exactMapper.index(*it)] = matrix[i][j];

 //std::cout << "elem: " << (int) exactMapper.map(*it) << "| (" << global[0] << ", " << global[1] << "), " << "(" << i+1 << ", " << j+1 << "), " << exactOscData[exactMapper.map(*it)] << std::endl;

        }

        // assemble stiffness matrix
        MatrixType exactBilinStiffMat, exactStiffMat, exactInterfaceMats;
        Assembler<DGBasis, DGBasis> exactAssembler(exactBasis, exactBasis);
        OscLocalAssembler<GridType, DGFE, DGFE> oscExactAssembler(exactOscData, exactMapper);
        exactAssembler.assembleOperator(oscExactAssembler, exactBilinStiffMat);

        //print(exactBilinStiffMat, "exactBilinStiffMat:");

        // assemble interface integrals
        MatrixIndexSet idxSet(exactBilinStiffMat.N(), exactBilinStiffMat.M());
        idxSet.exportIdx(exactInterfaceMats);


        IntersectionJumpMassAssembler<GV, DGFE, DGFE> localInterfaceAssembler(B);


        for (size_t i=0; i<exactFaultNetwork.size(); i++) {
            MatrixType exactInterfaceMat, res;
            const FaultInterface<GV>& fault = *(exactFaultNetwork.getInterface(i));

            InterfaceOperatorAssembler<DGBasis, DGBasis> exactInterfaceAssembler(exactBasis, exactBasis, fault);
            exactInterfaceAssembler.assemble(localInterfaceAssembler, exactInterfaceMat);

            //print(exactInterfaceMat, "exactInterfaceMat:");

            add(exactInterfaceMats, exactInterfaceMat, res);
            exactInterfaceMats = res;
        }

        add(exactBilinStiffMat, exactInterfaceMats, exactStiffMat);

        // assemble rhs
        Dune::BlockVector<Dune::FieldVector<double,1>> rhs, exactSol;
        rhs.resize(exactStiffMat.N());
        exactSol.resize(exactStiffMat.M());

        exactAssembler.assembleFunctional(l2FunctionalAssembler, rhs);

        // set boundary conditions
        BitVector boundaryDofs;
        BoundaryPatch<GridType::LevelGridView> boundaryPatch(exactGridView, true);
        constructBoundaryDofs(boundaryPatch, exactBasis, boundaryDofs);

        for(size_t i=0; i<boundaryDofs.size(); i++) {
            if(!boundaryDofs[i][0])
                continue;

            RowType& row = exactStiffMat[i];

            ColumnIterator cIt    = row.begin();
            ColumnIterator cEndIt = row.end();

            for(; cIt!=cEndIt; ++cIt) {
                row[cIt.index()] = 0;
            }

            row[i]=1;
            rhs[i] = 0;
            exactSol[i] = 0;
        }

        // compute exact solution

        timer.start();
        /*#if HAVE_UMFPACK
        Dune::UMFPack<MatrixType> exactSolver(exactStiffMat);
        exactSolver.apply(exactSol, rhs, res);
        #else
        #error No UMFPack!
        #endif*/

        #if HAVE_SUPERLU
        Dune::SuperLU<MatrixType> exactSolver(exactStiffMat, false);
        exactSolver.setVerbosity(false);
        exactSolver.apply(exactSol, rhs, res);
        #else
        #error No SuperLU!
        #endif

        std::cout << "Solving exact problem: " << timer.elapsed() << " seconds" << std::endl;
        timer.stop();
        timer.reset();


  /*  Dune::VTKWriter<GV> vtkwriter(exactGridView);
    vtkwriter.addVertexData(exactSol,"data");
    vtkwriter.write( "osccoeffImpl/exactvertexdata", Dune::VTK::appendedraw );*/

    EnergyNorm<MatrixType, Dune::BlockVector<Dune::FieldVector<ctype,1>> > exactEnergyNorm(exactStiffMat);


    // ----------------------------------------------------------------
    // ---                set up cg solver                          ---
    // ----------------------------------------------------------------
    std::cout << std::endl;
    std::cout << "---------------------------------------------" << std::endl; 
    std::cout << "Setting up the fine level CG solver:" << std::endl; 
    std::cout << "---------------------------------------------" << std::endl << std::endl;


    std::cout << "Assembling stiffness matrix" << std::endl;
    

    // assemble stiffness matrix
    MatrixType fineBilinStiffMat, fineStiffMat, fineInterfaceMats;
    Assembler<DGBasis, DGBasis> fineAssembler(fineBasis, fineBasis);
    OscLocalAssembler<GridType, DGFE, DGFE> oscFineAssembler(fineOscData, fineMapper);
    fineAssembler.assembleOperator(oscFineAssembler, fineBilinStiffMat);

    // assemble interface integrals
    MatrixIndexSet fineIdxSet(fineBilinStiffMat.N(), fineBilinStiffMat.M());
    fineIdxSet.exportIdx(fineInterfaceMats);

    for (size_t i=0; i<fineFaultNetwork.size(); i++) {
        MatrixType fineInterfaceMat, res;
        const FaultInterface<GV>& fault = *(fineFaultNetwork.getInterface(i));

        InterfaceOperatorAssembler<DGBasis, DGBasis> fineInterfaceAssembler(fineBasis, fineBasis, fault);
        fineInterfaceAssembler.assemble(localInterfaceAssembler, fineInterfaceMat);

        //print(fineInterfaceMat, "fineInterfaceMat:");

        add(fineInterfaceMats, fineInterfaceMat, res);
        fineInterfaceMats = res;
    }

    add(fineBilinStiffMat, fineInterfaceMats, fineStiffMat);

    std::cout << "Assembling right hand side" << std::endl;
    // assemble rhs
    Dune::BlockVector<Dune::FieldVector<double,1>> fineRhs, fineX;
    fineRhs.resize(fineStiffMat.N());
    fineX.resize(fineStiffMat.M());

    fineAssembler.assembleFunctional(l2FunctionalAssembler, fineRhs);
    
    std::cout << "Setting boundary conditions" << std::endl;
    // set boundary conditions
    BitVector fineBoundaryDofs;
    BoundaryPatch<GridType::LevelGridView> fineBoundaryPatch(fineGridView, true);
    constructBoundaryDofs(fineBoundaryPatch, fineBasis, fineBoundaryDofs);

    for(size_t i=0; i<fineBoundaryDofs.size(); i++) {
        if(!fineBoundaryDofs[i][0])
            continue;

        RowType& row = fineStiffMat[i];

        ColumnIterator cIt    = row.begin();
        ColumnIterator cEndIt = row.end();

        for(; cIt!=cEndIt; ++cIt) {
            row[cIt.index()] = 0;
        }

        row[i] = 1;
        fineRhs[i] = 0;
	fineX[i] = 0;
	initialX[i] = 0;
    }

    // TODO: remove after debugging
    /*typedef typename Dune::PQkLocalFiniteElementCache<ctype, double, dim, 1> P1FECache;
    typedef typename P1FECache::FiniteElementType FEType;
    P1FECache p1FECache;
    typedef typename GV::template Codim<0>::Iterator ElementIterator;

    ElementIterator fIt    = fineGridView.template begin<0>();
    ElementIterator fEndIt = fineGridView.template end<0>();

    for (; fIt != fEndIt; ++fIt) {
        int elemIdx = fineMapper.index(*fIt);

        const Dune::ReferenceElement<ctype,dim>& fineRefElement
            = Dune::ReferenceElements<ctype, dim>::general(fIt->type());

        // Get local finite element
        const FEType& fineBaseSet = p1FECache.get(fIt->type());

        const size_t numFineBaseFct = fineBaseSet.localBasis().size();

        std::vector<Dune::FieldVector<ctype,dim> > local(numFineBaseFct);

        for (size_t i=0; i<numFineBaseFct; i++)
        {
            const Dune::LocalKey& iLocalKey = fineBaseSet.localCoefficients().localKey(i);
            local[i] = fineRefElement.position(iLocalKey.subEntity(), iLocalKey.codim());
            int globalFine = fineBasis.index(*fIt, iLocalKey.subEntity());

            std::cout << "ElementIdx: " << elemIdx << "   " << globalFine << std::endl;
        }
    } */
    // end remove

    // compute exact solution on fine grid
    std::cout << "Computing the exact solution directly" << std::endl;
    #if HAVE_SUPERLU
    Dune::InverseOperatorResult opRes;
    Dune::BlockVector<Dune::FieldVector<double,1>> fineRhsClone(fineRhs);

    Dune::SuperLU<MatrixType> fineSolver(fineStiffMat);
    fineSolver.apply(fineX, fineRhsClone, opRes);
    #else
    #error No SuperLU!
    #endif

/*    Dune::VTKWriter<GV> fineVTKwriter(fineGridView);
    fineVTKwriter.addVertexData(fineX,"data");
    fineVTKwriter.write( "osccoeffImpl/finevertexdata", Dune::VTK::appendedraw );*/

    EnergyNorm<MatrixType, Dune::BlockVector<Dune::FieldVector<ctype,1>> > energyNorm(fineStiffMat);

    std::cout << "Setting up the preconditioner!" << std::endl;
    FaultPreconditioner<GridType, Basis, DGBasis, MatrixType, Dune::BlockVector<Dune::FieldVector<ctype,1>> > preconditioner(grid, 0, fineLevelIdx, fineOscData, fineMapper, fineStiffMat, fineRhs, fineFaultNetwork, false);

    std::cout << "Setup complete, starting preconditioned cg iteration!" << std::endl;
    std::cout << std::endl << std::endl;
    
/*    Dune::BlockVector<Dune::FieldVector<ctype,1>> initialError(fineX.size());
    initialError = initialX;
    initialError -= fineX;
    std::cout << "Initial error in energy norm: " << energyNorm.operator()(initialError) << std::endl;
    std::cout << "Initial error in two norm: " << initialError.two_norm() << std::endl << std::endl;*/

    Dune::BlockVector<Dune::FieldVector<ctype,1>> initialXCopy(initialX);
    Dune::BlockVector<Dune::FieldVector<ctype,1>> fineRhsCopy(fineRhs);


    //print(initialX, "initialX");
    //print(fineRhs, "fineRhs");
    //print(exactSol, "exactSol");
    //print(fineX, "fineX");
    
   /* Basis fineContBasis(fineGridView);
    std::cout << "fineContBasis.size(): " << fineContBasis.size() << std::endl;
    DGMGTransfer<Basis> contMGTransfer(coarseBasis, fineContBasis);

    Basis exactContBasis(exactGridView);
    std::cout << "exactContBasis.size(): " << exactContBasis.size() << std::endl; */
    DGMGTransfer<DGBasis> discMGTransfer(fineBasis, exactBasis);

    // solve
    OscCGSolver<MatrixType, Dune::BlockVector<Dune::FieldVector<ctype,1>>, DGBasis>
    solver(fineBasis, &fineStiffMat, &initialX, &fineRhs, &preconditioner,
                    maxIterations, solverTolerance, &exactEnergyNorm,
                    Solver::FULL, &exactSol, &fineX, discMGTransfer, 1.0, true); //((oscGridN+0.0)/fineGridN)
    solver.check();
    solver.preprocess();
    solver.solve();

    Dune::BlockVector<Dune::FieldVector<ctype,1>> finalError(fineX.size());
    finalError = initialX;
    finalError -= fineX;
    std::cout << "Final error in energy norm: " << energyNorm.operator()(finalError) << std::endl;
    std::cout << "Final error in two norm: " << finalError.two_norm() << std::endl << std::endl;

    std::cout << std::endl;
    std::cout << "Total time: " << totalTimer.elapsed() << " seconds." << std::endl; 

    // regular unpreconditioned cg solver for comparisson

    std::cout << "Setup complete, starting cg iteration!" << std::endl;
    std::cout << std::endl;
    
    // solve
    OscCGSolver<MatrixType, Dune::BlockVector<Dune::FieldVector<ctype,1>>, DGBasis >
    regSolver(fineBasis, &fineStiffMat, &initialXCopy, &fineRhsCopy, nullptr,
                    maxIterations, solverTolerance, &exactEnergyNorm,
                    Solver::FULL, &exactSol, &fineX, discMGTransfer, 1.0, true);
    regSolver.check();
    regSolver.preprocess();
    regSolver.solve();

    // TODO: use once debugging finished
    std::cout.rdbuf(coutbuf); //reset to standard output again

    std::cout << "Problem " << problemCount << " done!" << std::endl;
    problemCount++;
    }
} catch (Dune::Exception e) {

    std::cout << e << std::endl;

}
